"""
HTTP request related code.
"""

import posixpath
import re
import sys

import requests

from six.moves.urllib.parse import urlparse

from .exceptions import PyKubeError, HTTPError


_ipv4_re = re.compile(r"^((25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?).){3}(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$")


class HTTPClient(object):
    """
    Client for interfacing with the Kubernetes API.
    """

    def __init__(self, config):
        """
        Creates a new instance of the HTTPClient.

        :Parameters:
           - `config`: The configuration instance
        """
        self.config = config
        self.url = self.config.cluster["server"]
        self.session = self.build_session()

    @property
    def url(self):
        return self._url

    @url.setter
    def url(self, value):
        pr = urlparse(value)
        if sys.version_info < (3, 5) and ("::" in pr.hostname or _ipv4_re.match(pr.hostname)):
            raise PyKubeError("IP address hostnames are not supported with Python < 3.5. Please see https://github.com/kelproject/pykube/issues/29 for more info.")
        self._url = pr.geturl()

    def build_session(self):
        """
        Creates a new session for the client.
        """
        s = requests.Session()
        if "certificate-authority" in self.config.cluster:
            s.verify = self.config.cluster["certificate-authority"].filename()
        if "token" in self.config.user and self.config.user["token"]:
            s.headers["Authorization"] = "Bearer {}".format(self.config.user["token"])
        else:
            s.cert = (
                self.config.user["client-certificate"].filename(),
                self.config.user["client-key"].filename(),
            )
        return s

    def get_kwargs(self, **kwargs):
        """
        Creates a full URL to request based on arguments.

        :Parametes:
           - `kwargs`: All keyword arguments to build a kubernetes API endpoint
        """
        version = kwargs.pop("version", "v1")
        if version == "v1":
            base = kwargs.pop("base", "/api")
        elif any(map(version.startswith, ["extensions/", "batch/"])):
            base = kwargs.pop("base", "/apis")
        else:
            if "base" not in kwargs:
                raise TypeError("unknown API version; base kwarg must be specified.")
            base = kwargs.pop("base")
        bits = [base, version]
        if "namespace" in kwargs:
            namespace = kwargs.pop("namespace")
            if namespace:
                bits.extend([
                    "namespaces",
                    namespace,
                ])
        url = kwargs.get("url", "")
        if url.startswith("/"):
            url = url[1:]
        bits.append(url)
        kwargs["url"] = self.url + posixpath.join(*bits)
        return kwargs

    def raise_for_status(self, resp):
        try:
            resp.raise_for_status()
        except Exception:
            # attempt to provide a more specific exception based around what
            # Kubernetes returned as the error.
            if resp.headers["content-type"] == "application/json":
                payload = resp.json()
                if payload["kind"] == "Status":
                    raise HTTPError(payload["message"])
            raise

    def request(self, *args, **kwargs):
        """
        Makes an API request based on arguments.

        :Parameters:
           - `args`: Non-keyword arguments
           - `kwargs`: Keyword arguments
        """
        return self.session.request(*args, **self.get_kwargs(**kwargs))

    def get(self, *args, **kwargs):
        """
        Executes an HTTP GET.

        :Parameters:
           - `args`: Non-keyword arguments
           - `kwargs`: Keyword arguments
        """
        return self.session.get(*args, **self.get_kwargs(**kwargs))

    def options(self, *args, **kwargs):
        """
        Executes an HTTP OPTIONS.

        :Parameters:
           - `args`: Non-keyword arguments
           - `kwargs`: Keyword arguments
        """
        return self.session.options(*args, **self.get_kwargs(**kwargs))

    def head(self, *args, **kwargs):
        """
        Executes an HTTP HEAD.

        :Parameters:
           - `args`: Non-keyword arguments
           - `kwargs`: Keyword arguments
        """
        return self.session.head(*args, **self.get_kwargs(**kwargs))

    def post(self, *args, **kwargs):
        """
        Executes an HTTP POST.

        :Parameters:
           - `args`: Non-keyword arguments
           - `kwargs`: Keyword arguments
        """
        return self.session.post(*args, **self.get_kwargs(**kwargs))

    def put(self, *args, **kwargs):
        """
        Executes an HTTP PUT.

        :Parameters:
           - `args`: Non-keyword arguments
           - `kwargs`: Keyword arguments
        """
        return self.session.put(*args, **self.get_kwargs(**kwargs))

    def patch(self, *args, **kwargs):
        """
        Executes an HTTP PATCH.

        :Parameters:
           - `args`: Non-keyword arguments
           - `kwargs`: Keyword arguments
        """
        return self.session.patch(*args, **self.get_kwargs(**kwargs))

    def delete(self, *args, **kwargs):
        """
        Executes an HTTP DELETE.

        :Parameters:
           - `args`: Non-keyword arguments
           - `kwargs`: Keyword arguments
        """
        return self.session.delete(*args, **self.get_kwargs(**kwargs))



"""
pykube.config unittests
"""

import os

from pykube import config, exceptions

from . import TestCase


GOOD_CONFIG_FILE_PATH = os.path.sep.join(["test", "test_config.yaml"])


class TestConfig(TestCase):

    def setUp(self):
        self.cfg = config.KubeConfig.from_file(GOOD_CONFIG_FILE_PATH)

    def tearDown(self):
        self.cfg = None

    def test_init(self):
        """
        Test Config instance creation.
        """
        # Ensure that a valid creation works
        self.assertEqual(
            GOOD_CONFIG_FILE_PATH,
            self.cfg.filename)

        # Ensure that if a file does not exist the creation fails
        self.assertRaises(
            exceptions.PyKubeError,
            config.KubeConfig.from_file,
            "doesnotexist")

    def test_set_current_context(self):
        """
        Verify set_current_context works as expected.
        """
        self.cfg.set_current_context("new_context")
        self.assertEqual(
            "new_context",
            self.cfg.current_context)

    def test_clusters(self):
        """
        Verify clusters works as expected.
        """
        self.assertEqual(
            {"server": "http://localhost"},
            self.cfg.clusters.get("thecluster", None))

    def test_users(self):
        """
        Verify users works as expected.
        """
        self.assertEqual(
            "data",
            self.cfg.users.get("admin", None))

    def test_contexts(self):
        """
        Verify contexts works as expected.
        """
        self.assertEqual(
            {"cluster": "thecluster", "user": "admin"},
            self.cfg.contexts.get("thename", None))

    def test_cluster(self):
        """
        Verify cluster works as expected.
        """
        # Without a current_context this should fail
        try:
            cluster = self.cfg.cluster
            self.fail(
                "cluster was found without a current context set: {}".format(
                    cluster))
        except exceptions.PyKubeError:
            # We should get an error
            pass

        self.cfg.set_current_context("thename")
        self.assertEqual({"server": "http://localhost"}, self.cfg.cluster)

    def test_user(self):
        """
        Verify user works as expected.
        """
        # Without a current_context this should fail
        try:
            user = self.cfg.user
            self.fail(
                "user was found without a current context set: {}".format(
                    user))
        except exceptions.PyKubeError:
            # We should get an error
            pass

        self.cfg.set_current_context("thename")
        self.assertEqual("data", self.cfg.user)



"""
Python client for Kubernetes
"""

from .config import KubeConfig  # noqa
from .exceptions import KubernetesError, PyKubeError, ObjectDoesNotExist  # noqa
from .http import HTTPClient  # noqa
from .objects import (  # noqa
    ConfigMap,
    DaemonSet,
    Deployment,
    Endpoint,
    Ingress,
    Job,
    Namespace,
    Node,
    Pod,
    ReplicationController,
    ReplicaSet,
    Secret,
    Service,
)
from .query import now, all_ as all, everything  # noqa



"""
Configuration code.
"""

import base64
import copy
import tempfile
import os

import yaml

from pykube import exceptions


class KubeConfig(object):
    """
    Main configuration class.
    """

    @classmethod
    def from_service_account(cls, path="/var/run/secrets/kubernetes.io/serviceaccount"):
        with open(os.path.join(path, "token")) as fp:
            token = fp.read()
        host = os.environ.get("PYKUBE_KUBERNETES_SERVICE_HOST")
        if host is None:
            host = os.environ["KUBERNETES_SERVICE_HOST"]
        port = os.environ.get("PYKUBE_KUBERNETES_SERVICE_PORT")
        if port is None:
            port = os.environ["KUBERNETES_SERVICE_PORT"]
        doc = {
            "clusters": [
                {
                    "name": "self",
                    "cluster": {
                        "server": "https://{}:{}".format(host, port),
                        "certificate-authority": os.path.join(path, "ca.crt"),
                    },
                },
            ],
            "users": [
                {
                    "name": "self",
                    "user": {
                        "token": token,
                    },
                },
            ],
            "contexts": [
                {
                    "name": "self",
                    "context": {
                        "cluster": "self",
                        "user": "self",
                    },
                }
            ],
            "current-context": "self",
        }
        self = cls(doc)
        return self

    @classmethod
    def from_file(cls, filename):
        """
        Creates an instance of the KubeConfig class from a kubeconfig file.

        :Parameters:
           - `filename`: The full path to the configuration file
        """
        filename = os.path.expanduser(filename)
        if not os.path.isfile(filename):
            raise exceptions.PyKubeError("Configuration file {} not found".format(filename))
        with open(filename) as f:
            doc = yaml.safe_load(f.read())
        self = cls(doc)
        self.filename = filename
        return self

    def __init__(self, doc):
        """
        Creates an instance of the KubeConfig class.
        """
        self.doc = doc
        self.current_context = None
        if "current-context" in doc and doc["current-context"]:
            self.set_current_context(doc["current-context"])

    def set_current_context(self, value):
        """
        Sets the context to the provided value.

        :Parameters:
           - `value`: The value for the current context
        """
        self.current_context = value

    @property
    def clusters(self):
        """
        Returns known clusters by exposing as a read-only property.
        """
        if not hasattr(self, "_clusters"):
            cs = {}
            for cr in self.doc["clusters"]:
                cs[cr["name"]] = c = copy.deepcopy(cr["cluster"])
                if "server" not in c:
                    c["server"] = "http://localhost"
                BytesOrFile.maybe_set(c, "certificate-authority")
            self._clusters = cs
        return self._clusters

    @property
    def users(self):
        """
        Returns known users by exposing as a read-only property.
        """
        if not hasattr(self, "_users"):
            us = {}
            for ur in self.doc["users"]:
                us[ur["name"]] = u = copy.deepcopy(ur["user"])
                BytesOrFile.maybe_set(u, "client-certificate")
                BytesOrFile.maybe_set(u, "client-key")
            self._users = us
        return self._users

    @property
    def contexts(self):
        """
        Returns known contexts by exposing as a read-only property.
        """
        if not hasattr(self, "_contexts"):
            cs = {}
            for cr in self.doc["contexts"]:
                cs[cr["name"]] = copy.deepcopy(cr["context"])
            self._contexts = cs
        return self._contexts

    @property
    def cluster(self):
        """
        Returns the current selected cluster by exposing as a
        read-only property.
        """
        if self.current_context is None:
            raise exceptions.PyKubeError("current context not set; call set_current_context")
        return self.clusters[self.contexts[self.current_context]["cluster"]]

    @property
    def user(self):
        """
        Returns the current user by exposing as a read-only property.
        """
        if self.current_context is None:
            raise exceptions.PyKubeError("current context not set; call set_current_context")
        return self.users[self.contexts[self.current_context]["user"]]


class BytesOrFile(object):
    """
    Implements the same interface for files and byte input.
    """

    @classmethod
    def maybe_set(cls, d, key):
        file_key = key
        data_key = "{}-data".format(key)
        if data_key in d:
            d[file_key] = cls(d[data_key])
            del d[data_key]
        elif file_key in d:
            d[file_key] = cls(d[file_key])

    def __init__(self, data):
        """
        Creates a new instance of BytesOrFile.

        :Parameters:
           - `data`: A full path to a file or base64 encoded bytes
        """
        self._filename = None
        self._bytes = None
        if data.startswith("/"):
            self._filename = data
        else:
            self._bytes = base64.b64decode(data)

    def bytes(self):
        """
        Returns the provided data as bytes.
        """
        if self._filename:
            with open(self._filename, "rb") as f:
                return f.read()
        else:
            return self._bytes

    def filename(self):
        """
        Returns the provided data as a file location.
        """
        if self._filename:
            return self._filename
        else:
            with tempfile.NamedTemporaryFile(delete=False) as f:
                f.write(self._bytes)
            return f.name



import copy
import json
import time

import jsonpatch

from .exceptions import ObjectDoesNotExist
from .query import ObjectManager


DEFAULT_NAMESPACE = "default"


class APIObject(object):

    objects = ObjectManager()
    base = None
    namespace = None

    def __init__(self, api, obj):
        self.api = api
        self.set_obj(obj)

    def set_obj(self, obj):
        self.obj = obj
        self._original_obj = copy.deepcopy(obj)

    @property
    def name(self):
        return self.obj["metadata"]["name"]

    @property
    def annotations(self):
        return self.obj["metadata"].get("annotations", {})

    def api_kwargs(self, **kwargs):
        kw = {}
        collection = kwargs.pop("collection", False)
        if collection:
            kw["url"] = self.endpoint
        else:
            kw["url"] = "{}/{}".format(self.endpoint, self._original_obj["metadata"]["name"])
        if self.base:
            kw["base"] = self.base
        kw["version"] = self.version
        if self.namespace is not None:
            kw["namespace"] = self.namespace
        kw.update(kwargs)
        return kw

    def exists(self, ensure=False):
        r = self.api.get(**self.api_kwargs())
        if r.status_code not in {200, 404}:
            self.api.raise_for_status(r)
        if not r.ok:
            if ensure:
                raise ObjectDoesNotExist("{} does not exist.".format(self.name))
            else:
                return False
        return True

    def create(self):
        r = self.api.post(**self.api_kwargs(data=json.dumps(self.obj), collection=True))
        self.api.raise_for_status(r)
        self.set_obj(r.json())

    def reload(self):
        r = self.api.get(**self.api_kwargs())
        self.api.raise_for_status(r)
        self.set_obj(r.json())

    def update(self):
        patch = jsonpatch.make_patch(self._original_obj, self.obj)
        r = self.api.patch(**self.api_kwargs(
            headers={"Content-Type": "application/json-patch+json"},
            data=str(patch),
        ))
        self.api.raise_for_status(r)
        self.set_obj(r.json())

    def delete(self):
        r = self.api.delete(**self.api_kwargs())
        if r.status_code != 404:
            self.api.raise_for_status(r)


class NamespacedAPIObject(APIObject):

    objects = ObjectManager(namespace=DEFAULT_NAMESPACE)

    @property
    def namespace(self):
        if self.obj["metadata"].get("namespace"):
            return self.obj["metadata"]["namespace"]
        else:
            return DEFAULT_NAMESPACE


class ReplicatedAPIObject(object):

    @property
    def replicas(self):
        return self.obj["spec"]["replicas"]

    @replicas.setter
    def replicas(self, value):
        self.obj["spec"]["replicas"] = value


class ConfigMap(NamespacedAPIObject):

    version = "v1"
    endpoint = "configmaps"
    kind = "ConfigMap"


class DaemonSet(NamespacedAPIObject):

    version = "extensions/v1beta1"
    endpoint = "daemonsets"
    kind = "DaemonSet"


class Deployment(NamespacedAPIObject, ReplicatedAPIObject):

    version = "extensions/v1beta1"
    endpoint = "deployments"
    kind = "Deployment"


class Endpoint(NamespacedAPIObject):

    version = "v1"
    endpoint = "endpoints"
    kind = "Endpoint"


class Ingress(NamespacedAPIObject):

    version = "extensions/v1beta1"
    endpoint = "ingresses"
    kind = "Ingress"


class Job(NamespacedAPIObject):

    version = "batch/v1"
    endpoint = "jobs"
    kind = "Job"

    def scale(self, replicas=None):
        """
        Scales a job as it would be done by kubectl scale --replicas=num Jobs/myjob.
        Replicas can start from zero.
        """
        parallelism = replicas
        # we use parallelism from now on because this is what it is altered at
        # the API level by the kubectl call
        if parallelism is None:
            parallelism = self.obj["spec"]["parallelism"]
        self.exists(ensure=True)
        self.obj["spec"]["parallelism"] = parallelism
        self.update()
        while True:
            self.reload()
            if self.self.obj["spec"]["parallelism"] == parallelism:
                break
            time.sleep(1)


class Namespace(APIObject):

    version = "v1"
    endpoint = "namespaces"
    kind = "Namespace"


class Node(APIObject):

    version = "v1"
    endpoint = "nodes"
    kind = "Node"


class Pod(NamespacedAPIObject):

    version = "v1"
    endpoint = "pods"
    kind = "Pod"

    @property
    def ready(self):
        cs = self.obj["status"]["conditions"]
        condition = next((c for c in cs if c["type"] == "Ready"), None)
        return condition is not None and condition["status"] == "True"


class ReplicationController(NamespacedAPIObject, ReplicatedAPIObject):

    version = "v1"
    endpoint = "replicationcontrollers"
    kind = "ReplicationController"

    def scale(self, replicas=None):
        if replicas is None:
            replicas = self.replicas
        self.exists(ensure=True)
        self.replicas = replicas
        self.update()
        while True:
            self.reload()
            if self.replicas == replicas:
                break
            time.sleep(1)


class ReplicaSet(NamespacedAPIObject, ReplicatedAPIObject):

    version = "extensions/v1beta1"
    endpoint = "replicasets"
    kind = "ReplicaSet"


class Secret(NamespacedAPIObject):

    version = "v1"
    endpoint = "secrets"
    kind = "Secret"


class Service(NamespacedAPIObject):

    version = "v1"
    endpoint = "services"
    kind = "Service"


class PersistentVolume(APIObject):

    version = "v1"
    endpoint = "persistentvolumes"
    kind = "PersistentVolume"


class PersistentVolumeClaim(NamespacedAPIObject):

    version = "v1"
    endpoint = "persistentvolumeclaims"
    kind = "PersistentVolumeClaim"



from setuptools import setup, find_packages


with open("README.rst") as fp:
    long_description = fp.read()


setup(
    name="pykube",
    version="0.11.3",
    description="Python client library for Kubernetes",
    long_description=long_description,
    author="Eldarion, Inc.",
    author_email="development@eldarion.com",
    license="Apache",
    url="https://github.com/eldarion-gondor/pykube",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
    ],
    zip_safe=False,
    packages=find_packages(),
    install_requires=[
        "requests",
        "PyYAML",
        "six",
        "jsonpatch",
    ],
)



import logging
import math
import time

from .objects import Pod
from .exceptions import KubernetesError


logger = logging.getLogger(__name__)


class RollingUpdater(object):

    def __init__(self, api, old_rc, new_rc, **kwargs):
        self.api = api
        self.old_rc = old_rc
        self.new_rc = new_rc
        self.update_period = kwargs.get("update_period", 10)
        self.max_unavailable = kwargs.get("max_unavailable", 0)
        self.max_surge = kwargs.get("max_surge", 1)

    def update(self):
        desired = self.new_rc.replicas
        original = self.old_rc.replicas
        max_unavailable = extract_max_value(self.max_unavailable, "max_unavailable", desired)
        max_surge = extract_max_value(self.max_surge, "max_surge", desired)
        min_available = original - max_unavailable
        if self.new_rc.exists():
            logger.info("ReplicationController {} already exists.".format(self.new_rc.name))
            return False
        new_selector = self.new_rc.obj["spec"]["selector"]
        old_selector = self.old_rc.obj["spec"]["selector"]
        if new_selector == old_selector:
            raise KubernetesError(
                "error: {} must specify a matching key with non-equal value in Selector for {}".format(
                    self.new_rc.name,
                    self.old_rc.name
                ))
        new_labels = self.new_rc.obj["spec"]["template"]["metadata"]["labels"]
        if new_selector != new_labels:
            raise KubernetesError(
                "The ReplicationController {} is invalid. spec.template.metadata.labels: Invalid value: {}: `selector` does not match template `labels` {}".format(
                    self.new_rc.name,
                    new_selector,
                    new_labels))

        self.create_rc(self.new_rc)
        logger.info("Created {}".format(self.new_rc.name))
        new_rc, old_rc = self.new_rc, self.old_rc

        logger.info(
            "scaling up {} from {} to {}, scaling down {} from {} to 0 (keep {} pods available, don't exceed {} pods)".format(
                new_rc.name,
                new_rc.replicas,
                desired,
                old_rc.name,
                old_rc.replicas,
                min_available,
                original + max_surge
            ),
        )

        while new_rc.replicas != desired or old_rc.replicas != 0:
            scaled_rc = self.scale_up(
                new_rc, old_rc,
                original, desired,
                max_surge, max_unavailable,
            )
            new_rc = scaled_rc
            time.sleep(self.update_period)
            scaled_rc = self.scale_down(
                new_rc, old_rc,
                desired,
                min_available, max_surge,
            )
            old_rc = scaled_rc

        logger.info("Update succeeded. Deleting {}".format(old_rc.name))
        self.cleanup(old_rc, new_rc)

    def scale_up(self, new_rc, old_rc, original, desired, max_surge, max_unavailable):
        # if we're already at the desired, do nothing.
        if new_rc.replicas == desired:
            return new_rc
        # scale up as far as we can based on the surge limit.
        increment = (original + max_surge) - (old_rc.replicas + new_rc.replicas)
        # if the old is already scaled down, go ahead and scale all the way up.
        if old_rc.replicas == 0:
            increment = desired - new_rc.replicas
        # we can't scale up without violating the surge limit, so do nothing
        if increment <= 0:
            return new_rc
        # increase the replica count, and deal with fenceposts
        new_rc.replicas = min(desired, new_rc.replicas + increment)
        # perform the scale up
        logger.info("scaling {} up to {}".format(new_rc.name, new_rc.replicas))
        new_rc.scale()
        return new_rc

    def scale_down(self, new_rc, old_rc, desired, min_available, max_surge):
        # already scaled down; do nothing.
        if old_rc.replicas == 0:
            return old_rc
        # block until there are any pods ready
        _, new_available = self.poll_for_ready_pods(old_rc, new_rc)
        # the old controller is considered as part of the total because we want
        # to maintain minimum availability even with a volatile old controller.
        # scale down as much as possible while maintaining minimum availability.
        decrement = old_rc.replicas + new_available - min_available
        # the decrement normally shouldn't drop below zero because the available
        # count always start below the old replica count, but the old replica
        # count can decrement due to externalities like pods death in the replica
        # set. this will be considered a transient condition; do nothing and try
        # again later with new readiness values.
        #
        # if the most we can scale is zero, it means we can't scale down without
        # violating the minimum. do nothing and try again later when conditions
        # may have changed.
        if decrement <= 0:
            return old_rc
        # reduce the replica count, and deal with fenceposts
        old_rc.replicas = max(0, old_rc.replicas - decrement)
        # if the new is already fully scaled and available up to the desired size,
        # go ahead and scale old all the way down
        if new_rc.replicas == desired and new_available == desired:
            old_rc.replicas = 0
        # perform scale down
        logger.info("scaling {} down to {}".format(old_rc.name, old_rc.replicas))
        old_rc.scale()
        return old_rc

    def cleanup(self, old_rc, new_rc):
        old_rc.delete()

    def poll_for_ready_pods(self, old_rc, new_rc):
        controllers = [old_rc, new_rc]
        old_ready = 0
        new_ready = 0
        any_ready = False

        while True:
            for controller in controllers:
                pods = Pod.objects(self.api).filter(
                    namespace=controller.namespace,
                    selector=controller.obj["spec"]["selector"],
                )
                for pod in pods:
                    if pod.ready:
                        if controller.name == old_rc.name:
                            old_ready += 1
                        elif controller.name == new_rc.name:
                            new_ready += 1
                        any_ready = True
            if any_ready:
                break
            time.sleep(1)

        return old_ready, new_ready

    def create_rc(self, rc):
        rc.replicas = 0
        rc.create()


def extract_max_value(field, name, value):
    assert type(field) in {int, str}, "{} is not an int or str".format(type(field))
    if isinstance(field, int):
        assert field >= 0, "{} must be >= 0".format(name)
        return field
    if isinstance(field, str):
        v = int(field.replace("%", ""))
        assert v >= 0, "{} must be >= 0".format(name)
        return math.ceil(float(value) * (float(v) / 100.))
